#include <time.h>
#include <vector>
#include <stdlib.h>
#include <math.h>

#include "util.h"
#include "sparse.h"

using namespace std;

const double ZERO_IN_CRS = 0.000001;

//      CRS (3 ,   )
//  
//   
double Transpose1(crsMatrix imtx, crsMatrix &omtx)
{
  clock_t start, finish;

  start = clock();

  vector<int>* columns = new vector<int>[imtx.N];
  vector<double> *values = new vector<double>[imtx.N];

  int count = 0;
  for (int i = 0; i < imtx.N; i++)
    for (int j = imtx.RowIndex[i]; j < imtx.RowIndex[i + 1]; j++)
    {
      columns[imtx.Col[j]].push_back(i);
      values[imtx.Col[j]].push_back(imtx.Value[j]);
      count++;
    }

  InitializeMatrix(imtx.N, count, omtx);

  count = 0;
  for (int i = 0; i < imtx.N; i++)
    for (unsigned int j = 0; j < columns[i].size(); j++)
    {
      omtx.Col[count] = columns[i][j];
      omtx.Value[count] = values[i][j];
      count++;
    }

  omtx.RowIndex[0] = 0;
  int sum = int(columns[0].size());
  for (int i = 1; i < omtx.N; i++)
  {
    omtx.RowIndex[i] = sum;
    sum += int(columns[i].size());
  }
  omtx.RowIndex[omtx.N] = omtx.NZ;

  finish = clock();

  delete [] columns;
  delete [] values;

  return double(finish - start) / CLOCKS_PER_SEC;
}

double Transpose2(crsMatrix imtx, crsMatrix &omtx)
{
  clock_t start, finish;
  int i, j;

  start = clock();

  InitializeMatrix(imtx.N, imtx.NZ, omtx);

  memset(omtx.RowIndex, 0, (imtx.N + 1) * sizeof(int));
  for (i = 0; i < imtx.NZ; i++) 
    omtx.RowIndex[imtx.Col[i] + 1]++;
  
  int S = 0;
  for (i = 1; i <= imtx.N; i++) 
  {
    int tmp = omtx.RowIndex[i];
    omtx.RowIndex[i] = S;
    S = S + tmp;
  }

  for (i = 0; i < imtx.N; i++) 
  {
    int j1 = imtx.RowIndex[i];
    int j2 = imtx.RowIndex[i+1];
    int Col = i; //   AT -   
    for (j = j1; j < j2; j++) 
    {
      double V = imtx.Value[j];  // 
      int RIndex = imtx.Col[j];  //   AT
      int IIndex = omtx.RowIndex[RIndex + 1];
      omtx.Value[IIndex] = V;
      omtx.Col  [IIndex] = Col;
      omtx.RowIndex[RIndex + 1]++;
    }
  }

  finish = clock();

  return double(finish - start) / CLOCKS_PER_SEC;
}

//  2     CRS (3 ,   )
//  C = A * B, C -   CRS (3 ,   )
//                         C     
//    : 0 - , 1 -    (N)
//   
int Multiplicate(crsMatrix A, crsMatrix B, crsMatrix &C, double &time)
{
  if (A.N != B.N)
    return 1;

  int N = A.N;
  vector<int> columns;
  vector<double> values;
  vector<int> row_index;

  clock_t start = clock();
  int rowNZ;
  row_index.push_back(0);
  for (int i = 0; i < N; i++)
  {
    rowNZ = 0;
    for (int j = 0; j < N; j++)
    {
      double sum = 0;
      for (int k = A.RowIndex[i]; k < A.RowIndex[i + 1]; k++)
      {
        for (int l = B.RowIndex[j]; l < B.RowIndex[j + 1]; l++)
        {
          if (A.Col[k] == B.Col[l])
          {
            sum += A.Value[k] * B.Value[l];
            break;
          }
        }
      }
      if (fabs(sum) > ZERO_IN_CRS)
      {
        columns.push_back(j);
        values.push_back(sum);
        rowNZ++;
      }
    }
    row_index.push_back(rowNZ + row_index[i]);
  }

  InitializeMatrix(N, columns.size(), C);

  for (unsigned int j = 0; j < columns.size(); j++)
  {
    C.Col[j] = columns[j];
    C.Value[j] = values[j];
  }
  for(int i = 0; i <= N; i++)
    C.RowIndex[i] = row_index[i];

  clock_t finish = clock();

  time = (double)(finish - start) / CLOCKS_PER_SEC;

  return 0;
}

