#include <math.h>
#include <memory.h>
#include <stdlib.h>
#include <time.h>
#include <vector>
#include "mkl_spblas.h"
#include "util.h"

using namespace std;

const double MAX_VAL = 100.0;

//  -      
bool isSrandCalled = false;

//      CRS (3 ,   )
//     Value, Col  RowIndex
//    mtx
void InitializeMatrix(int N, int NZ, crsMatrix &mtx)
{
  mtx.N = N;
  mtx.NZ = NZ;
  mtx.Value    = new double[NZ];
  mtx.Col      = new int[NZ];
  mtx.RowIndex = new int[N + 1];
}

//  ,    mtx
void FreeMatrix(crsMatrix &mtx)
{
  delete [] mtx.Value;
  delete [] mtx.Col;
  delete [] mtx.RowIndex;
}

//   imtx  omtx,     Value, Col  RowIndex
void CopyMatrix(crsMatrix imtx, crsMatrix &omtx)
{
  //   
  int N  = imtx.N;
  int NZ = imtx.NZ;
  InitializeMatrix(N, NZ, omtx);
  // 
  memcpy(omtx.Value   , imtx.Value   , NZ * sizeof(double));
  memcpy(omtx.Col     , imtx.Col     , NZ * sizeof(int));
  memcpy(omtx.RowIndex, imtx.RowIndex, (N + 1) * sizeof(int));
}

//  2     CRS (3 ,   )
//  max|Cij|,  C = A - B
//    : 0 - , 1 -    (N)
int CompareMatrix(crsMatrix mtx1, crsMatrix mtx2, double &diff)
{
  int N = mtx1.N;
  int i, j;
  int i1, i2;
  double **p;

  //   ?
  if (mtx1.N != mtx2.N)
    return 1; //   

  N = mtx1.N;
  //   
  p = new double*[N];
  for (i = 0; i < N; i++) 
  {
    p[i] = new double[N];  
    for (j = 0; j < N; j++)
      p[i][j] = 0.0;
  }

  //     
  for (i = 0; i < N; i++)
  {
    i1 = mtx1.RowIndex[i];
    i2 = mtx1.RowIndex[i + 1] - 1;
    for (j = i1; j <= i2; j++)
      p[i][mtx1.Col[j]] = mtx1.Value[j];
  }

  //     
  for (i = 0; i < N; i++)
  {
    i1 = mtx2.RowIndex[i];
    i2 = mtx2.RowIndex[i + 1] - 1;
    for (j = i1; j <= i2; j++)
      p[i][mtx2.Col[j]] -= mtx2.Value[j];
  }

  //   
  double max = 0.0;
  for (i = 0; i < N; i++)
    for (j = 0; j < N; j++)
      if (fabs(p[i][j]) > max)
        max = fabs(p[i][j]);
  diff = max;

  //   
  for (i = 0; i < N; i++)
    delete[] p[i];
  delete[] p;

  return 0; //  
}

//  2     CRS (3 ,   )
//  max|Cij|,  C = A - B
//      MKL
//    : 0 - , 1 -    (N)
int SparseDiff(crsMatrix A, crsMatrix B, double &diff)
{
  if (A.N != B.N)
    return 1;

  int n = A.N;

  //   C = A - B,  MKL
  //     MKL
  double *c = 0; // 
  int *jc = 0;   //   (  )
  int *ic;   //     (  )
  
  //      MKL
  //   A  B  
  int i, j;
  for (i = 0; i < A.NZ; i++)
    A.Col[i]++;
  for (i = 0; i < B.NZ; i++)
    B.Col[i]++;
  for (j = 0; j <= A.N; j++)
  {
    A.RowIndex[j]++;
    B.RowIndex[j]++;
  }

  //  ,  C = A + beta*op(B)
  char trans = 'N'; //   , op(B) = B -    B

//  ,   ,    
// request = 0:     ..  
//    ,      ,
// :
// 1)       ic: "- +1" ;
// 2)     request = 1 -   ic   
//                                                          
// 3)     c  jc 
//    (-  = ic[- ]-1)
// 4)     request = 2
  int request;

//    :   ,   
//   A, B  C.   ,   
// , ,   "No-No-Yes", 
//   ,     1  7 
  int sort = 8;

// beta = -1 -> C = A + (-1.0) * B;
  double beta = -1.0;

//   .
//    request = 0
  int nzmax = -1;

//  
  int info;

//       C
  ic = new int[n + 1];
//       C
  request = 1;
  mkl_dcsradd(&trans, &request, &sort, &n, &n, A.Value, A.Col, A.RowIndex, 
    &beta, B.Value, B.Col, B.RowIndex, c, jc, ic, &nzmax, &info);
  int nzc = ic[n] - 1;
  c = new double[nzc];
  jc = new int[nzc];
//  C = A - B
  request = 2;
  mkl_dcsradd(&trans, &request, &sort, &n, &n, A.Value, A.Col, A.RowIndex, 
    &beta, B.Value, B.Col, B.RowIndex, c, jc, ic, &nzmax, &info);
//  max|Cij|
  diff = 0.0;
  for (i = 0; i < nzc; i++)
  {
    double var = fabs(c[i]);
    if (var > diff)
      diff = var;
  }
//      A  B
  for (i = 0; i < A.NZ; i++)
    A.Col[i]--;
  for (i = 0; i < B.NZ; i++)
    B.Col[i]--;
  for (j = 0; j <= n; j++)
  {
    A.RowIndex[j]--;
    B.RowIndex[j]--;
  }
//  
  delete [] c;
  delete [] ic;
  delete [] jc;

  return 0;
}

//  2     CRS (3 ,   )
//  C = A * B, C -   CRS (3 ,   )
//                         C     
//    : 0 - , 1 -    (N)
//   
int SparseMKLMult(crsMatrix A, crsMatrix B, crsMatrix &C, double &time)
{
  if (A.N != B.N)
    return 1;

  int n = A.N;

  //      MKL
  //   A  B  
  int i, j;
  for (i = 0; i < A.NZ; i++)
    A.Col[i]++;
  for (i = 0; i < B.NZ; i++)
    B.Col[i]++;
  for (j = 0; j <= n; j++)
  {
    A.RowIndex[j]++;
    B.RowIndex[j]++;
  }

  //  ,  C = op(A) * B
  char trans = 'N'; //   , op(A) = A -    A

//  ,   ,    
// request = 0:     ..  
//    ,      ,
// :
// 1)       ic: "- +1" ;
// 2)     request = 1 -   ic   
//                                                          
// 3)     c  jc 
//    (-  = ic[- ]-1)
// 4)     request = 2
  int request;

//    :   ,   
//   A, B  C.   ,   
// , ,   "No-No-Yes", 
//   ,     1  7 
  int sort = 8;

//   .
//    request = 0
  int nzmax = -1;

//  
  int info;

  clock_t start = clock();

//       C
  C.RowIndex = new int[n + 1];
//       C
  request = 1;
  C.Value = 0;
  C.Col = 0;
  mkl_dcsrmultcsr(&trans, &request, &sort, &n, &n, &n, A.Value, A.Col, 
                  A.RowIndex, B.Value, B.Col, B.RowIndex, C.Value, C.Col,
                  C.RowIndex, &nzmax, &info);
  int nzc = C.RowIndex[n] - 1;
  C.Value = new double[nzc];
  C.Col = new int[nzc];
//  C = A * B
  request = 2;
  mkl_dcsrmultcsr(&trans, &request, &sort, &n, &n, &n, A.Value, A.Col,
                  A.RowIndex, B.Value, B.Col, B.RowIndex, C.Value, C.Col,
                  C.RowIndex, &nzmax, &info);
  C.N = n;
  C.NZ = nzc;

  clock_t finish = clock();

  //      A, B  
  for (i = 0; i < A.NZ; i++)
    A.Col[i]--;
  for (i = 0; i < B.NZ; i++)
    B.Col[i]--;
  for (i = 0; i < C.NZ; i++)
    C.Col[i]--;
  for (j = 0; j <= n; j++)
  {
    A.RowIndex[j]--;
    B.RowIndex[j]--;
    C.RowIndex[j]--;
  }

  time = double(finish - start) / double(CLOCKS_PER_SEC);

  return 0;
}

double next()
{
  return ((double)rand() / (double)RAND_MAX);
}

//      CRS (3 ,   )
//    cntInRow  
void GenerateRegularCRS(int seed, int N, int cntInRow, crsMatrix& mtx)
{
  int i, j, k, f, tmp, notNull, c;

  if (!isSrandCalled)
  {
    srand(seed);
    isSrandCalled = true;
  }

  notNull = cntInRow * N;
  InitializeMatrix(N, notNull, mtx);

  for(i = 0; i < N; i++)
  {
    //      i
    for(j = 0; j < cntInRow; j++)
    {
      do
      {
        mtx.Col[i * cntInRow + j] = rand() % N;
        f = 0;
        for (k = 0; k < j; k++)
          if (mtx.Col[i * cntInRow + j] == mtx.Col[i * cntInRow + k])
            f = 1;
      } while (f == 1);
    }
    //      i
    for (j = 0; j < cntInRow - 1; j++)
      for (k = 0; k < cntInRow - 1; k++)
        if (mtx.Col[i * cntInRow + k] > mtx.Col[i * cntInRow + k + 1])
        {
          tmp = mtx.Col[i * cntInRow + k];
          mtx.Col[i * cntInRow + k] = mtx.Col[i * cntInRow + k + 1];
          mtx.Col[i * cntInRow + k + 1] = tmp;
        }
  }

  //   
  for (i = 0; i < cntInRow * N; i++)
    mtx.Value[i] = next() * MAX_VAL;

  //    
  c = 0;
  for (i = 0; i <= N; i++)
  {
    mtx.RowIndex[i] = c;
    c += cntInRow;
  }
}

//      CRS (3 ,   )
//        1  cntInRow
//   -  
void GenerateSpecialCRS(int seed, int N, int cntInRow, crsMatrix& mtx)
{
  if (!isSrandCalled)
  {
    srand(seed);
    isSrandCalled = true;
  }

  double end = pow((double)cntInRow, 1.0 / 3.0);
  double step = end / N;

  vector<int>* columns = new vector<int>[N];
  int NZ = 0;

  for (int i = 0; i < N; i++)
  {
    int rowNZ = int(pow((double(i + 1) * step), 3) + 1);
    NZ += rowNZ;
    int num1 = (rowNZ - 1) / 2;
    int num2 = rowNZ - 1 - num1;

    if (rowNZ != 0)
    {
      if (i < num1)
      {
        num2 += num1 - i;
        num1 = i;
        for(int j = 0; j < i; j++)
          columns[i].push_back(j);
        columns[i].push_back(i);
        for(int j = 0; j < num2; j++)
          columns[i].push_back(i + 1 + j);

      }
      else
      {
        if (N - i - 1 < num2)
        {
          num1 += num2 - (N - 1 - i);
          num2 = N - i - 1;
        }
        for (int j = 0; j < num1; j++)
          columns[i].push_back(i - num1 + j);
        columns[i].push_back(i);
        for (int j = 0; j < num2; j++)
          columns[i].push_back(i + j + 1);
      }
    }
  }

  InitializeMatrix(N, NZ, mtx);

  int count = 0;
  int sum = 0;
  for (int i = 0; i < N; i++)
  {
    mtx.RowIndex[i] = sum;
    sum += columns[i].size();
    for (unsigned int j = 0; j < columns[i].size(); j++)
    {
      mtx.Col[count] = columns[i][j];
      mtx.Value[count] = next();
      count++;
    }
  }
  mtx.RowIndex[N] = sum;

  delete [] columns;
}
