#include <time.h>
#include <vector>
#include <stdlib.h>
#include <math.h>

#include "util.h"
#include "sparse.h"

using namespace std;

const double ZERO_IN_CRS = 0.000001;

//      CRS (3 ,   )
//  
//   
double Transpose1(crsMatrix imtx, crsMatrix &omtx)
{
  clock_t start, finish;

  start = clock();

  vector<int>* columns = new vector<int>[imtx.N];
  vector<double> *values = new vector<double>[imtx.N];

  int count = 0;
  for (int i = 0; i < imtx.N; i++)
    for (int j = imtx.RowIndex[i]; j < imtx.RowIndex[i + 1]; j++)
    {
      columns[imtx.Col[j]].push_back(i);
      values[imtx.Col[j]].push_back(imtx.Value[j]);
      count++;
    }

  InitializeMatrix(imtx.N, count, omtx);

  count = 0;
  for (int i = 0; i < imtx.N; i++)
    for (unsigned int j = 0; j < columns[i].size(); j++)
    {
      omtx.Col[count] = columns[i][j];
      omtx.Value[count] = values[i][j];
      count++;
    }

  omtx.RowIndex[0] = 0;
  int sum = int(columns[0].size());
  for (int i = 1; i < omtx.N; i++)
  {
    omtx.RowIndex[i] = sum;
    sum += int(columns[i].size());
  }
  omtx.RowIndex[omtx.N] = omtx.NZ;

  finish = clock();

  delete [] columns;
  delete [] values;

  return double(finish - start) / CLOCKS_PER_SEC;
}

double Transpose2(crsMatrix imtx, crsMatrix &omtx)
{
  clock_t start, finish;
  int i, j;

  start = clock();

  InitializeMatrix(imtx.N, imtx.NZ, omtx);

  memset(omtx.RowIndex, 0, (imtx.N + 1) * sizeof(int));
  for (i = 0; i < imtx.NZ; i++) 
    omtx.RowIndex[imtx.Col[i] + 1]++;
  
  int S = 0;
  for (i = 1; i <= imtx.N; i++) 
  {
    int tmp = omtx.RowIndex[i];
    omtx.RowIndex[i] = S;
    S = S + tmp;
  }

  for (i = 0; i < imtx.N; i++) 
  {
    int j1 = imtx.RowIndex[i];
    int j2 = imtx.RowIndex[i+1];
    int Col = i; //   AT -   
    for (j = j1; j < j2; j++) 
    {
      double V = imtx.Value[j];  // 
      int RIndex = imtx.Col[j];  //   AT
      int IIndex = omtx.RowIndex[RIndex + 1];
      omtx.Value[IIndex] = V;
      omtx.Col  [IIndex] = Col;
      omtx.RowIndex[RIndex + 1]++;
    }
  }

  finish = clock();

  return double(finish - start) / CLOCKS_PER_SEC;
}

//  2     CRS (3 ,   )
//  C = A * B, C -   CRS (3 ,   )
//     C     
//    : 0 - , 1 -    (N)
//   
int Multiplicate(crsMatrix A, crsMatrix B, crsMatrix &C, double &time)
{
  if (A.N != B.N)
    return 1;

  int N = A.N;
  int i, j, k;
  int NZ = 0;

  vector<int> columns;
  vector<double> values;
  vector<int> row_index;

  clock_t start = clock();
  
  int *temp = new int[N];
  row_index.push_back(0);

  for (i = 0; i < N; i++)
  {
    // i-   A
    //     
    memset(temp, -1, N * sizeof(int));
    //         
    int ind1 = A.RowIndex[i], ind2 = A.RowIndex[i + 1];
    for (j = ind1; j < ind2; j++)
    {
      int col = A.Col[j];
      temp[col] = j; // ,  a[i, ]  
                     //    Value   temp[]
    }
    //    i  A
    //          BT
    for (j = 0; j < N; j++)
    {
      // j-   B
      double sum = 0;
      int ind3 = B.RowIndex[j], ind4 = B.RowIndex[j + 1];
      //     j  B
      for (k = ind3; k < ind4; k++)
      {
        int bcol = B.Col[k];
        int aind = temp[bcol];
        if (aind != -1)
          sum += A.Value[aind] * B.Value[k];
      }
      if (fabs(sum) > ZERO_IN_CRS)
      {
        columns.push_back(j);
        values.push_back(sum);
        NZ++;
      }
    }
    row_index.push_back(NZ);
  }

  InitializeMatrix(N, NZ, C);

  for (j = 0; j < NZ; j++)
  {
    C.Col[j] = columns[j];
    C.Value[j] = values[j];
  }
  for(i = 0; i <= N; i++)
    C.RowIndex[i] = row_index[i];

  delete [] temp;

  clock_t finish = clock();
  time = (double)(finish - start) / CLOCKS_PER_SEC;

  return 0;
}

//  2     CRS (3 ,   )
//  C = A * B, C -   CRS (3 ,   )
//     C     
//      
//    : 0 - , 1 -    (N)
//   
int SymbolicMult(crsMatrix A, crsMatrix B, crsMatrix &C, double &time)
{
  if (A.N != B.N)
    return 1;

  int N = A.N;
  int i, j, k;
  int NZ = 0;

  vector<int> columns;
  vector<int> row_index;

  clock_t start = clock();

  int *temp = new int[N];
  row_index.push_back(0);
  for (i = 0; i < N; i++)
  {
    // i-   A
    //     
    memset(temp, -1, N * sizeof(int));
    //         
    int ind1 = A.RowIndex[i], ind2 = A.RowIndex[i + 1];
    for (j = ind1; j < ind2; j++)
    {
      int col = A.Col[j];
      temp[col] = j; // ,  a[i, ]  
                     //    Value   temp[]
    }
    //    i  A
    //          BT
    for (j = 0; j < N; j++)
    {
      // j-   B
      int IsFound = 0;
      int ind3 = B.RowIndex[j], ind4 = B.RowIndex[j + 1];
      //     j  B
      for (k = ind3; k < ind4; k++)
      {
        int bcol = B.Col[k];
        int aind = temp[bcol];
        if (aind != -1)
        {
          IsFound = 1;
          break;
        }
      }
      if (IsFound)
      {
        columns.push_back(j);
        NZ++;
      }
    }
    row_index.push_back(NZ);
  }

  InitializeMatrix(N, NZ, C);

  for (int j = 0; j < columns.size(); j++)
  {
    C.Col[j] = columns[j];
  }
  for (int i = 0; i <= N; i++)
    C.RowIndex[i] = row_index[i];

  delete [] temp;

  clock_t finish = clock();

  time = (double)(finish - start) / CLOCKS_PER_SEC;

  return 0;
}

//  2     CRS (3 ,   )
//  C = A * B, C -   CRS (3 ,   )
//      
//    C     
//    : 0 - , 1 -    (N)
//   
int NumericMult(crsMatrix A, crsMatrix B, crsMatrix &C, double &time)
{
  if (A.N != B.N)
    return 1;

  int i, j, k, l;
  int N = A.N;

  clock_t start = clock();
  memset(C.Value, 0, C.NZ * sizeof(double));

  int index = 0;
  for (i = 0; i < N; i++)
  {
    for (j = C.RowIndex[i]; j < C.RowIndex[i + 1]; j++)
    {
      for (k = A.RowIndex[i]; k < A.RowIndex[i + 1]; k++)
      {
        int colJ = C.Col[j];
        for (l = B.RowIndex[colJ]; l < B.RowIndex[colJ + 1]; l++)
        {
          if (A.Col[k] == B.Col[l])
          {
            C.Value[index] += A.Value[k] * B.Value[l];
            break;
          }
        }
      }
      index++;
    }
  }

  clock_t finish = clock();

  time = (double)(finish - start) / CLOCKS_PER_SEC;

  return 0;
}

int NumericMult1(crsMatrix A, crsMatrix B, crsMatrix &C, double &time)
{
  if (A.N != B.N)
    return 1;

  int N = A.N;

  clock_t start = clock();
  memset(C.Value, 0, C.NZ * sizeof(double));

  int index = 0;
  for (int i = 0; i < N; i++)
  {
    for (int j = C.RowIndex[i]; j < C.RowIndex[i + 1]; j++)
    {
      //   i  A   j  B
      double sum = 0;
      int colJ = C.Col[j];
      int ks = A.RowIndex[i]; int ls = B.RowIndex[colJ]; 
      int kf = A.RowIndex[i + 1] - 1; int lf = B.RowIndex[colJ + 1] - 1;
      while ((ks <= kf) && (ls <= lf))
      {
        if (A.Col[ks] < B.Col[ls])
          ks++;
        else 
          if (A.Col[ks] > B.Col[ls])
            ls++;
          else 
          {
            sum += A.Value[ks] * B.Value[ls];
            ks++;
            ls++;
          }
      }
      C.Value[index] = sum;
      index++;
    }
  }

  clock_t finish = clock();

  time = (double)(finish - start) / CLOCKS_PER_SEC;

  return 0;
}

int NumericMult2(crsMatrix A, crsMatrix B, crsMatrix &C, double &time)
{
  if (A.N != B.N)
    return 1;

  int N = A.N;

  clock_t start = clock();
  int *temp = new int[N];
  int index = 0;

  for (int i = 0; i < N; i++)
  {
    int ind1 = C.RowIndex[i];
    int ind2 = C.RowIndex[i + 1];
    if (ind2 > ind1)
    {
      memset(temp, -1, sizeof(int) * N);
      int Aind1 = A.RowIndex[i];
      int Aind2 = A.RowIndex[i + 1];
      for (int j = Aind1; j < Aind2; j++)
      {
        int col = A.Col[j];
        temp[col] = j; // ,  a[i, ]  
                     //    Value   temp[]
      }
      //  ,//   i  A    B
      for (int j = ind1; j < ind2; j++)
      {
        int col = C.Col[j];
        //   i-      col  BT
        double sum = 0;
        int Bind1 = B.RowIndex[col];
        int Bind2 = B.RowIndex[col + 1];
        for (int k = Bind1; k < Bind2; k++)
        {
          int lookup = B.Col[k];
          int aind = temp[lookup];
          if (aind != -1)
            sum += A.Value[aind] * B.Value[k];
        }
        C.Value[index] = sum;
        index++;
      }
    }
  }

  delete [] temp;

  clock_t finish = clock();

  time = (double)(finish - start) / CLOCKS_PER_SEC;

  return 0;
}
