#include <time.h>
#include <vector>
#include <stdlib.h>
#include <math.h>

#include "util.h"
#include "sparse.h"

#include "omp.h"

using namespace std;

const double ZERO_IN_CRS = 0.000001;

//      CRS (3 ,   )
//  
//   
double Transpose1(crsMatrix imtx, crsMatrix &omtx)
{
  clock_t start, finish;

  start = clock();

  vector<int>* columns = new vector<int>[imtx.N];
  vector<double> *values = new vector<double>[imtx.N];

  int count = 0;
  for (int i = 0; i < imtx.N; i++)
    for (int j = imtx.RowIndex[i]; j < imtx.RowIndex[i + 1]; j++)
    {
      columns[imtx.Col[j]].push_back(i);
      values[imtx.Col[j]].push_back(imtx.Value[j]);
      count++;
    }

  InitializeMatrix(imtx.N, count, omtx);

  count = 0;
  for (int i = 0; i < imtx.N; i++)
    for (unsigned int j = 0; j < columns[i].size(); j++)
    {
      omtx.Col[count] = columns[i][j];
      omtx.Value[count] = values[i][j];
      count++;
    }

  omtx.RowIndex[0] = 0;
  int sum = int(columns[0].size());
  for (int i = 1; i < omtx.N; i++)
  {
    omtx.RowIndex[i] = sum;
    sum += int(columns[i].size());
  }
  omtx.RowIndex[omtx.N] = omtx.NZ;

  finish = clock();

  delete [] columns;
  delete [] values;

  return double(finish - start) / CLOCKS_PER_SEC;
}

double Transpose2(crsMatrix imtx, crsMatrix &omtx)
{
  clock_t start, finish;
  int i, j;

  start = clock();

  InitializeMatrix(imtx.N, imtx.NZ, omtx);

  memset(omtx.RowIndex, 0, (imtx.N + 1) * sizeof(int));
  for (i = 0; i < imtx.NZ; i++) 
    omtx.RowIndex[imtx.Col[i] + 1]++;
  
  int S = 0;
  for (i = 1; i <= imtx.N; i++) 
  {
    int tmp = omtx.RowIndex[i];
    omtx.RowIndex[i] = S;
    S = S + tmp;
  }

  for (i = 0; i < imtx.N; i++) 
  {
    int j1 = imtx.RowIndex[i];
    int j2 = imtx.RowIndex[i+1];
    int Col = i; //   AT -   
    for (j = j1; j < j2; j++) 
    {
      double V = imtx.Value[j];  // 
      int RIndex = imtx.Col[j];  //   AT
      int IIndex = omtx.RowIndex[RIndex + 1];
      omtx.Value[IIndex] = V;
      omtx.Col  [IIndex] = Col;
      omtx.RowIndex[RIndex + 1]++;
    }
  }

  finish = clock();

  return double(finish - start) / CLOCKS_PER_SEC;
}

//  2     CRS (3 ,   )
//  C = A * B, C -   CRS (3 ,   )
//     C     
//    : 0 - , 1 -    (N)
//   
int Multiplicate(crsMatrix A, crsMatrix B, crsMatrix &C, double &time)
{
  if (A.N != B.N)
    return 1;

  int N = A.N;
  int i, j, k;

  clock_t start = clock();

  vector<int>* columns = new vector<int>[N];
  vector<double> *values = new vector<double>[N];

  int* row_index = new int[N + 1];
  memset(row_index, 0, sizeof(int) * N);

  int chunk = 10;
#pragma omp parallel
  {
    int *temp = new int[N];
  #pragma omp for private(j, k) schedule(static, chunk)
    for (i = 0; i < N; i++)
    {
      // i-   A
      //     
      memset(temp, -1, N * sizeof(int));
      //         
      int ind1 = A.RowIndex[i], ind2 = A.RowIndex[i + 1];
      for (j = ind1; j < ind2; j++)
      {
        int col = A.Col[j];
        temp[col] = j; // ,  a[i, ]  
        //    Value   temp[]
      }
      //    i  A
      //          BT
      for (j = 0; j < N; j++)
      {
        // j-   B
        double sum = 0;
        int ind3 = B.RowIndex[j], ind4 = B.RowIndex[j + 1];
        //     j  B
        for (k = ind3; k < ind4; k++)
        {
          int bcol = B.Col[k];
          int aind = temp[bcol];
          if (aind != -1)
            sum += A.Value[aind] * B.Value[k];
        }
        if (fabs(sum) > ZERO_IN_CRS)
        {
          columns[i].push_back(j);
          values[i].push_back(sum);
          row_index[i]++;
        }
      }
    }
    delete [] temp;
  }

  int NZ = 0;
  for(i = 0; i < N; i++)
  {
    int tmp = row_index[i];
    row_index[i] = NZ;
    NZ += tmp;
  }
  row_index[N] = NZ;

  InitializeMatrix(N, NZ, C);

  int count = 0;
  for (i = 0; i < N; i++)
  {
    int size = columns[i].size();
    memcpy(&C.Col[count], &columns[i][0], size * sizeof(int));
    memcpy(&C.Value[count], &values[i][0], size * sizeof(double));
    count += size;
  }
  memcpy(C.RowIndex, &row_index[0], (N + 1) * sizeof(int));

  delete [] row_index;
  delete [] columns;
  delete [] values;

  clock_t finish = clock();
  time = (double)(finish - start) / CLOCKS_PER_SEC;

  return 0;
}
