#ifndef __SPARSE_H__
#define __SPARSE_H__

#include <vector>
#include <math.h>

#include "tbb/task_scheduler_init.h"
#include "tbb/blocked_range.h"
#include "tbb/parallel_for.h"

using namespace std;
using namespace tbb;

const double ZERO_IN_CRS = 0.000001;

struct crsMatrix
{
  int N;  //   (N x N)
  int NZ; // -  

  //   ( NZ)
  double* Value;
  //    ( NZ)
  int* Col;
  //    ( N + 1)
  int* RowIndex; 
};

class Multiplicator
{
  crsMatrix A, B;
  vector<int>* columns;
  vector<double>* values;
  int *row_index;
public:
  Multiplicator(crsMatrix& _A, crsMatrix& _B, vector<int>* &_columns, 
    vector<double>* &_values, int *_row_index) :
    A(_A), B(_B), columns(_columns), values(_values), row_index(_row_index)
  {}

  void operator()(const blocked_range<int>& r) const
  {
    int begin = r.begin();
    int end = r.end();
    int N = A.N;

    int i, j, k;
    int *temp = new int[N];

    for (i = begin; i < end; i++)
    {
      // i-   A
      //     
      memset(temp, -1, N * sizeof(int));
      //         
      int ind1 = A.RowIndex[i], ind2 = A.RowIndex[i + 1];
      for (j = ind1; j < ind2; j++)
      {
        int col = A.Col[j];
        temp[col] = j; // ,  a[i, ]  
        //    Value   temp[]
      }
      //    i  A
      //          BT
      for (j = 0; j < N; j++)
      {
        // j-   B
        double sum = 0;
        int ind3 = B.RowIndex[j], ind4 = B.RowIndex[j + 1];
        //     j  B
        for (k = ind3; k < ind4; k++)
        {
          int bcol = B.Col[k];
          int aind = temp[bcol];
          if (aind != -1)
            sum += A.Value[aind] * B.Value[k];
        }
        if (fabs(sum) > ZERO_IN_CRS)
        {
          columns[i].push_back(j);
          values[i].push_back(sum);
          row_index[i]++;
        }
      }
    }
    delete [] temp;
  }
};

//      CRS (3 ,   )
//  
//   
double Transpose1(crsMatrix imtx, crsMatrix &omtx);
double Transpose2(crsMatrix imtx, crsMatrix &omtx);

//  2     CRS (3 ,   )
//  C = A * B, C -   CRS (3 ,   )
//     C     
//    : 0 - , 1 -    (N)
//   
int Multiplicate(crsMatrix A, crsMatrix B, crsMatrix &C, double &time);

#endif // __SPARSE_H__